!******************************************************************************!
!                                                                              !
!    process_definition_rcl.f90                                                !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016,2017 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!

module process_definition_rcl

!------------------------------------------------------------------------------!

  use globals_rcl
  use order_rcl, only: init_max_couplings, max_couplings_init
  use modelfile, only: get_particle_id_mdl, get_antiparticle_id_mdl,   &
                       get_n_orders_mdl, get_order_id_mdl

!------------------------------------------------------------------------------!

  implicit none

!------------------------------------------------------------------------------!

  contains

!------------------------------------------------------------------------------!

  subroutine define_process_rcl(npr,processIn,order)

  integer,          intent(in) :: npr
  character(len=*), intent(in) :: processIn,order


  integer                :: i,j,n,n0,n1,n2,le,leIn,l0,lmax,l1,sign, &
                            nd0,nd,pond(99),binnd(99),polegs,nord

  integer, allocatable   :: inprT(:),legsInT(:),legsOutT(:),             &
                            parT(:,:),helT(:,:),resMaxT(:),              &
                            binResT(:,:),parResT(:,:),powselT(:,:,:,:),  &
                            pa(:),paT(:),he(:),heT(:)
                            
  logical, allocatable   :: loopT(:)
  integer, allocatable   :: qflowT(:,:)
  character              :: cpr*99,cpr0*99,che*3
  character, allocatable :: cpa(:)*7,cpaT(:)*7,processT(:)*99

  call processes_generated_error_rcl('define_process_rcl')
  if (.not. max_couplings_init) then
    call init_max_couplings
    max_couplings_init = .true.
  end if

  cpr = adjustl(processIn)

  n = 99
  do while (index (cpr,' [').ne.0)
    n = index (cpr,' [')
    cpr = cpr(1:n-1)//cpr(n+1:)
  enddo
  n = 99
  do while (scan(cpr,'(').ne.0)
    n = scan(cpr,'(')
    cpr = cpr(1:n-1)//' { '//cpr(n+1:)
  enddo
  n = 99
  do while (scan(cpr,')').ne.0)
    n = scan(cpr,')')
    cpr = cpr(1:n-1)//' } '//cpr(n+1:)
  enddo

  le = 0
  leIn = 0

  pond = 0
  nd0 = 0
  nd = 0

  nord = get_n_orders_mdl()

 do while (cpr.ne.'')

    if (le.gt.0) then
      allocate (cpaT(le)); cpaT(:le) = cpa(:le); deallocate (cpa)
      allocate ( paT(le));  paT(:le) =  pa(:le); deallocate ( pa)
      allocate ( heT(le));  heT(:le) =  he(:le); deallocate ( he)
    endif
    le = le + 1
    allocate (cpa(le),pa(le),he(le))
    if (le.gt.1) then
      cpa(:le-1) = cpaT(:le-1); deallocate (cpaT)
       pa(:le-1) =  paT(:le-1); deallocate ( paT)
       he(:le-1) =  heT(:le-1); deallocate ( heT)
    endif

    n0 = scan(cpr,'>')
    n1 = scan(cpr,' ') - 1
    sign = 1
    if (n1.ge.n0) then
      cpr = adjustl(cpr(n0+1:))
      n1 = scan(cpr,' ') - 1
      sign = -1
    else
      leIn = le
    endif

    n2 = scan(cpr(1:n1),'[')
    if (n2.eq.0) then
      cpa(le) = cpr(1:n1);        he(le) = 111
    else
      che = cpr(n2:n2+2)
      if     (che.eq.'[-]') then; he(le) = -1 * sign
      elseif (che.eq.'[0]') then; he(le) =  0
      elseif (che.eq.'[+]') then; he(le) = +1 * sign
      else;                       he(le) = 111
      endif
      cpa(le) = cpr(1:n2-1)
    endif

    cpr0 = adjustl(cpr(n1+1:))
    if (cpr0(1:1).eq.'{'.and.he(le).ne.111) then
      call warning_rcl('Particle ' // trim(cpa(le)),    &
                       ' can not have a specific helicity.')
      he(le) = 111
    endif

    call openOutput
    pa(le) = get_particle_id_mdl(cpa(le))
    if (sign.eq.-1) pa(le) = get_antiparticle_id_mdl(cpa(le))
    cpr = adjustl(cpr(n1+1:))
    if (cpr(1:1).eq.'{') then
      nd0 = nd0 + 1
      nd  = nd0
      pond(nd) = pa(le)
      l1 = le
      le = le - 1
      cpr = adjustl(cpr(2:))
    elseif (cpr(1:1).eq.'}') then
      polegs = le - l1 + 1
      cpr = adjustl(cpr(2:))
      binnd(nd) = 2**(l1-1)*( 2**polegs - 1 )
      nd = nd - 1
    endif

  enddo

  l0 = 0
  if (prTot.gt.0) then
    l0 = maxval(legsIn+legsOut)
    allocate (processT(         prTot));      processT = process
    allocate (   inprT(         prTot));         inprT =    inpr
    allocate ( legsInT(         prTot));       legsInT =  legsIn
    allocate (legsOutT(         prTot));      legsOutT = legsOut
    allocate (    parT(l0,      prTot));          parT =     par
    allocate (    helT(l0,      prTot));          helT =     hel
    allocate ( resMaxT(         prTot));       resMaxT =  resMax
    allocate ( binResT(l0,      prTot));       binResT =  binRes
    allocate ( parResT(l0,      prTot));       parResT =  parRes
    allocate (   loopT(         prTot));         loopT =    loop
    allocate (  qflowT(l0,      prTot));        qflowT =    qflow
    allocate (powselT(0:2*l0,nord,0:1,prTot)); powselT =   powsel
    deallocate (process,inpr,legsIn,legsOut,par,hel,resMax,binRes, &
                parRes,loop,qflow,powsel)
  endif

  prTot = prTot + 1

  lmax = max(l0,le)

  allocate (process(           prTot))
  allocate (   inpr(           prTot));    inpr = 0
  allocate ( legsIn(           prTot));  legsIn = 0
  allocate (legsOut(           prTot)); legsOut = 0
  allocate (    par(lmax,      prTot));     par = 0
  allocate (    hel(lmax,      prTot));     hel = 0
  allocate ( resMax(           prTot));  resMax = 0
  allocate ( binRes(lmax,      prTot));  binRes = 0
  allocate ( parRes(lmax,      prTot));  parRes = 0
  allocate (   loop(           prTot));    loop = .false.
  allocate (  qflow(lmax,      prTot));   qflow = 0

  ! Increased size of powersel to support dim8 operators with order LAM=2 
  allocate (powsel(0:2*lmax,nord,0:1,prTot)); powsel = 1

  if (prTot.gt.1) then
    process(         :prTot-1) = processT
       inpr(         :prTot-1) =    inprT
     legsIn(         :prTot-1) =  legsInT
    legsOut(         :prTot-1) = legsOutT
        par(:l0,     :prTot-1) =     parT
        hel(:l0,     :prTot-1) =     helT
     resMax(         :prTot-1) =  resMaxT
     binRes(:l0,     :prTot-1) =  binResT
     parRes(:l0,     :prTot-1) =  parResT
       loop(         :prTot-1) =    loopT
      qflow(:l0,     :prTot-1) =   qflowT
    powsel(0:2*l0,:nord,0:1,:prTot-1) = powselT
    deallocate (processT,inprT,legsInT,legsOutT,parT,helT,resMaxT, &
                binResT,parResT,loopT,qflowT,powselT)
  endif

  process(prTot) = adjustl(processIn)

  inpr(prTot) = npr
  do i = 1,prTot
  do j = 1,prTot
    if (i.ne.j.and.inpr(i).eq.inpr(j)) then
      call error_rcl('define_process_rcl called for different '// &
                     'processes with the same process-index')
    endif
  enddo
  enddo

  if (leIn.le.0) then
    call error_rcl('define_process_rcl called for a process ' // &
                   'without incoming particles')
  endif

   legsIn(      prTot) = leIn
  legsOut(      prTot) = le - leIn
      par(:le,  prTot) = pa
      hel(:le,  prTot) = he
   resMax(      prTot) = nd0
   binRes(1:nd0,prTot) = binnd(1:nd0)
   parRes(1:nd0,prTot) = pond(1:nd0)

  if     (order.eq.'LO' ) then; loop(prTot) = .false.
  elseif (order.eq.'NLO') then; loop(prTot) = .true.; loopMax = .true.
  else
    call error_rcl('define_process_rcl called at the wrong '// &
                   "loop order (accepted values are order = 'LO','NLO')")
  endif

  legsMax = maxval(legsIn+legsOut)

  end subroutine define_process_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! The next subroutines select/unselect the contributions to the 
! amplitude with specific powers of g_s (i.e. the strong coupling: 
! g_s^2 = 4*pi*alpha_s).
! If none of them are called all contributions are selected 
! (default). 
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_ord(cid,funcname,ord)
    ! returns the order id "ord" given the fundamental order string "cid".  In
    ! case the order id is not found an exception is raised naming the original
    ! function "funcname" calling get_ord.
    character(len=*),  intent(in) :: cid,funcname
    integer,          intent(out) :: ord

    integer        :: nord,i
    character(100) :: order_ids

    ! check fundamental order
    nord = get_n_orders_mdl()
    ord = -1
    do i = 1, nord
      if (get_order_id_mdl(i) .eq. cid) then
        ord = i
        exit
      end if
    end do

    if (ord .eq. -1) then
      order_ids = ' '
      do i = 1, n_orders
        order_ids = trim(order_ids)//' '//trim(adjustl(get_order_id_mdl(i)))
      end do
      call error_rcl('Undefined order id. Allowed values ' //&
                     adjustl(order_ids), where=funcname)
    endif
  end subroutine get_ord

!------------------------------------------------------------------------------!

  subroutine select_power_BornAmpl_rcl(npr,cid,power)
    ! This subroutine selects for the process with process number :f:var:`npr`
    ! the contribution to the Born amplitude with power :f:var:`power` in the
    ! fundamental order :f:var:`cid`. The process needs to be defined via
    ! :f:subr:`define_process_rcl` before calling this subroutine.
    ! The selection of the contributions to the loop amplitude are uneffected,
    ! as well as the selection of the other powers for the Born amplitude.
    ! For SM-like theories with the only two fundamental coupling types (:f:var:`cid`) ’QCD’
    ! and ’QED’ the methods :f:subr:`select_gs_power_BornAmpl_rcl` and 
    ! :f:subr:`unselect_gs_power_BornAmpl_rcl` can be used instead.

    integer,          intent(in) :: npr    ! Process identifier
    integer,          intent(in) :: power  ! Power :m:`n` in :ref:`fundamental coupling<fd>` 
    character(len=*), intent(in) :: cid    ! String identifier for :ref:`fundamental coupling<fd>`(e.g. 'QED', 'QCD')

    integer :: pr,legs,ord

    call processes_generated_warning_rcl('select_power_BornAmpl_rcl')

    ! get internal process index
    call get_pr(npr, 'select_power_BornAmpl_rcl', pr)
    
    ! check power
    legs = legsIn(pr) + legsOut(pr)
    if (power .lt. 0 .or. power .gt. legs-2) then
      call warning_rcl('select_power_BornAmpl_rcl called with a possibly ' // &
                       'invalid power and might have no effect. ' // &
                       'The power should be between 0 and number of legs-2.')
    endif

    ! get internal fundamental id
    call get_ord(cid,'select_power_BornAmpl_rcl',ord)

    powsel(power,ord,0,pr) = 1

  end subroutine select_power_BornAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine unselect_power_BornAmpl_rcl(npr,cid,power)
    ! This subroutine unselects for the process with process number :f:var:`npr`
    ! the contribution to the Born amplitude with power :f:var:`power` in the
    ! fundamental order :f:var:`cid`. The process needs to be defined via
    ! :f:subr:`define_process_rcl` before calling this subroutine.
    ! The selection of the contributions to the loop amplitude are uneffected,
    ! as well as the selection of the other powers for the Born amplitude.
    ! For SM-like theories with the only two fundamental coupling types (:f:var:`cid`) ’QCD’
    ! and ’QED’ the methods :f:subr:`select_gs_power_BornAmpl_rcl` and 
    ! :f:subr:`unselect_gs_power_BornAmpl_rcl` can be used instead.

    integer,          intent(in) :: npr    ! Process identifier
    integer,          intent(in) :: power  ! Power :m:`n` in :ref:`fundamental coupling<fd>` 
    character(len=*), intent(in) :: cid    ! String identifier for :ref:`fundamental coupling<fd>`(e.g. 'QED', 'QCD')

    integer :: pr,legs,ord

    call processes_generated_warning_rcl('unselect_power_BornAmpl_rcl')

    ! get internal process index
    call get_pr(npr, 'unselect_all_powers_BornAmpl_rcl', pr)

    ! check power
    legs = legsIn(pr) + legsOut(pr)
    if (power .lt. 0 .or. power .gt. legs-2) then
      call warning_rcl('unselect_power_BornAmpl_rcl called with a possibly ' // &
                       'invalid power and might have no effect. ' //&
                       'The power should be between 0 and number of legs-2.')
    endif

    ! get internal fundamental id
    call get_ord(cid, 'unselect_power_BornAmpl_rcl', ord)

    powsel(power,ord,0,pr) = 0

  end subroutine unselect_power_BornAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine select_all_powers_BornAmpl_rcl(npr)

    ! This subroutine selects for the process with process number "npr" 
    ! all contribution to the Born amplitude (with any g_s power). 
    ! The selection of the contributions to the loop amplitude remains  
    ! uneffected. 

    integer, intent(in) :: npr
    integer             :: pr

    call processes_generated_warning_rcl('select_all_powers_BornAmpl_rcl')

    ! get internal process index
    call get_pr(npr,'select_all_powers_BornAmpl_rcl',pr)

    powsel(0:,:,0,pr) = 1

  end subroutine select_all_powers_BornAmpl_rcl

!------------------------------------------------------------------------------!
  
  subroutine unselect_all_powers_BornAmpl_rcl(npr)

    ! This subroutine unselects for the process with process number "npr" all
    ! contribution to the Born amplitude. The selection of the contributions to
    ! the loop amplitude remains uneffected.

    integer, intent(in) :: npr
    integer             :: pr

    call processes_generated_warning_rcl('unselect_all_powers_BornAmpl_rcl')

    ! get internal process index
    call get_pr(npr, 'unselect_all_powers_BornAmpl_rcl', pr)

    powsel(0:,:,0,pr) = 0

  end subroutine unselect_all_powers_BornAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine select_power_LoopAmpl_rcl(npr,cid,power)

    ! This subroutine selects for the process with process number "npr" the
    ! contribution to the Loop amplitude with power "power" in the fundamental
    ! order "cid".
    ! The selection of the contributions to the loop amplitude are 
    ! uneffected, as well as the selection of the other powers of g_s 
    ! for the Born amplitude.

    integer,          intent(in) :: npr,power
    character(len=*), intent(in) :: cid

    integer :: pr,legs,ord

    call processes_generated_warning_rcl('select_power_LoopAmpl_rcl')

    ! get internal process index
    call get_pr(npr, 'select_power_LoopAmpl_rcl', pr)

    ! check power
    legs = legsIn(pr) + legsOut(pr)
    if (power .lt. 0 .or. power .gt. legs) then
      call warning_rcl('select_power_LoopAmpl_rcl called with a possibly ' // &
                       'invalid power and might have no effect. ' //&
                       'The power should be between 0 and number of legs.')
    endif

    ! get internal fundamental id
    call get_ord(cid,'select_power_LoopAmpl_rcl',ord)

    powsel(power,ord,1,pr) = 1

  end subroutine select_power_LoopAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine unselect_power_LoopAmpl_rcl(npr,cid,power)

    ! This subroutine unselects for the process with process number 
    ! "npr" the contribution to the loop amplitude with g_s power 
    ! "gspower".
    ! The selection of the contributions to the Born amplitude remains 
    ! uneffected, as well as the selection of the other powers of g_s 
    ! for the loop amplitude.

    integer,          intent(in) :: npr,power
    character(len=*), intent(in) :: cid

    integer :: pr,legs,ord

    call processes_generated_warning_rcl('unselect_power_LoopAmpl_rcl')

    ! get internal process index
    call get_pr(npr,'unselect_power_LoopAmpl_rcl',pr)

    ! check power
    legs = legsIn(pr) + legsOut(pr)
    if (power .lt. 0 .or. power.gt.legs) then
      call warning_rcl('unselect_power_LoopAmpl_rcl called with a possibly ' // &
                       'invalid power and might have no effect. ' //&
                       'The power should be between 0 and number of legs.')
    endif

    ! get internal fundamental id
    call get_ord(cid,'unselect_power_LoopAmpl_rcl',ord)

    powsel(power,ord,1,pr) = 0

  end subroutine unselect_power_LoopAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine select_all_powers_LoopAmpl_rcl(npr)

    ! This subroutine selects for the process with process number "npr" 
    ! all contribution to the Loop amplitude (with any power). 
    ! The selection of the contributions to the Born amplitude remains  
    ! uneffected. 

    integer, intent(in) :: npr
    integer             :: pr

    call processes_generated_warning_rcl('select_all_powers_LoopAmpl_rcl')

    ! get internal process index
    call get_pr(npr,'select_all_powers_LoopAmpl_rcl',pr)

    powsel(0:,:,1,pr) = 1

  end subroutine select_all_powers_LoopAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine unselect_all_powers_LoopAmpl_rcl(npr)

    ! This subroutine unselects for the process with process number 
    ! "npr" all contribution to the loop amplitude (with any g_s 
    ! power). 
    ! The selection of the contributions to the Born amplitude remains 
    ! uneffected. 

    integer, intent(in) :: npr
    integer             :: pr

    call processes_generated_warning_rcl('unselect_all_powers_LoopAmpl_rcl')

    ! get internal process index
    call get_pr(npr,'unselect_all_powers_LoopAmpl_rcl',pr)

    powsel(0:,:,1,pr) = 0

  end subroutine unselect_all_powers_LoopAmpl_rcl

!------------------------------------------------------------------------------!

  subroutine split_collier_cache_rcl(npr,n)

    ! This subroutine splits the cache of collier for process "npr"
    ! in "n" parts.

    integer, intent(in) :: npr,n
    integer             :: pr

    call processes_generated_warning_rcl('set_collier_cache_rcl')

    ! get internal process index
    call get_pr(npr,'set_collier_cache_rcl',pr)

    if (n.lt.0) then
      call error_rcl('set_collier_cache_rcl called with wrong arguments')
    endif

    if (.not.allocated(nCache)) then
      allocate(nCache(1:prTot)); nCache = 1
    endif

    if (.not.allocated(cacheOn)) then
      allocate(cacheOn(1:prTot)); cacheOn = .true.
    endif

    if (n.eq.0) then
      cacheOn(pr) = .false.
    else
      nCache(pr) = n
    endif

  end subroutine split_collier_cache_rcl

!------------------------------------------------------------------------------!

end module process_definition_rcl

!------------------------------------------------------------------------------!
